@@ -1,3 +1,4 @@
+#
 # Copyright 2013 The py-lmdb authors, all rights reserved.
 #
 # Redistribution and use in source and binary forms, with or without
@@ -17,6 +18,7 @@
 #
 # Additional information about OpenLDAP can be obtained at
 # <http://www.openldap.org/>.
+#
 
 """
 CPython/cffi wrapper for OpenLDAP's "Lightning" MDB database.
@@ -27,10 +29,8 @@ Please see http://lmdb.readthedocs.org/
 from __future__ import absolute_import
 
 import os
-import shutil
-import tempfile
+import sys
 import threading
-import warnings
 import weakref
 
 import lmdb
@@ -50,6 +50,12 @@ __all__ += ['Error', 'KeyExistsError', 'NotFoundError', 'PageNotFoundError',
             'BadTxnError', 'BadValsizeError', 'ReadonlyError',
             'InvalidParameterError', 'LockError', 'MemoryError', 'DiskError']
 
+
+# Handle moronic Python >=3.0 <3.3.
+UnicodeType = type('')
+if UnicodeType is bytes:
+    UnicodeType = str
+
 # Used to track context across cffi callbcks.
 _callbacks = threading.local()
 
@@ -177,6 +183,7 @@ _CFFI_CDEF = '''
     #define MDB_READERS_FULL ...
     #define MDB_REVERSEKEY ...
     #define MDB_TXN_FULL ...
+    #define MDB_NORDAHEAD ...
     #define MDB_WRITEMAP ...
     #define MDB_NOTLS ...
 
@@ -261,6 +268,7 @@ if not lmdb._reading_docs():
     # Try to use distutils-bundled cffi configuration to avoid a recompile and
     # potential compile errors during first module import.
     _config_vars = _config.CONFIG if _config else {
+        'extra_compile_args': ['-w'],
         'extra_sources': ['lib/mdb.c', 'lib/midl.c'],
         'extra_include_dirs': ['lib'],
         'extra_library_dirs': [],
@@ -273,7 +281,7 @@ if not lmdb._reading_docs():
         modulename='lmdb_cffi',
         ext_package='lmdb',
         sources=_config_vars['extra_sources'],
-        extra_compile_args=['-Wno-shorten-64-to-32'],
+        extra_compile_args=_config_vars['extra_compile_args'],
         include_dirs=_config_vars['extra_include_dirs'],
         libraries=_config_vars['libraries'],
         library_dirs=_config_vars['extra_library_dirs'])
@@ -290,7 +298,7 @@ if not lmdb._reading_docs():
     def _msg_func(s, _):
         """mdb_msg_func() callback. Appends `s` to _callbacks.msg_func list.
         """
-        _callbacks.msg_func.append(_ffi.string(s))
+        _callbacks.msg_func.append(_ffi.string(s).decode())
         return 0
 
 class Error(Exception):
@@ -410,7 +418,7 @@ class DiskError(Error):
 
 # Prepare _error_map, a mapping of integer MDB_ERROR_CODE to exception class.
 _error_map = {}
-for obj in globals().values():
+for obj in list(globals().values()):
     if getattr(obj, '__name__', '').endswith('Error'):
         code = globals().get(getattr(obj, 'MDB_NAME', None))
         _error_map[code] = obj
@@ -433,12 +441,22 @@ def _kill_dependents(parent):
 
 class Some_LMDB_Resource_That_Was_Deleted_Or_Closed(object):
     """We need this because cffi on PyPy treats None as cffi.NULL, instead of
-    throwing an exception it feeds MDB null pointers. We use a weird name to
-    make exceptions more obvious."""
+    throwing an exception it feeds LMDB null pointers. That means simply
+    replacing native handles with None during _kill_dependents() will lead to
+    NULL pointer derefences. Instead we use this class, and its weird name to
+    cause a TypeError, with a very obvious string appearing in the exception
+    text.
+
+    The only alternatives to this are inserting a check around every single use
+    of a native handle to ensure the handle is still valid prior to calling
+    LMDB, or doing no crash-safety checking at all.
+    """
     def __nonzero__(self):
         return 0
+    def __bool__(self):
+        return False
     def __repr__(self):
-        return "<This used a resource that was deleted or closed>"
+        return "<This used to be a LMDB resource but it was deleted or closed>"
 _invalid = Some_LMDB_Resource_That_Was_Deleted_Or_Closed()
 
 def _depend(parent, child):
@@ -539,6 +557,11 @@ class Environment(object):
         `create`:
             If ``False``, do not create the directory `path` if it is missing.
 
+        `readahead`:
+            If ``False``, LMDB will disable the OS filesystem readahead
+            mechanism, which may improve random read performance when a
+            database is larger than RAM.
+
         `writemap`:
             If ``True`` LMDB will use a writeable memory map to update the
             database. This option is incompatible with nested transactions.
@@ -584,8 +607,8 @@ class Environment(object):
     """
     def __init__(self, path, map_size=10485760, subdir=True,
             readonly=False, metasync=True, sync=True, map_async=False,
-            mode=0o644, create=True, writemap=False, max_readers=126,
-            max_dbs=0, max_spare_txns=1, max_spare_cursors=32,
+            mode=0o644, create=True, readahead=True, writemap=False,
+            max_readers=126, max_dbs=0, max_spare_txns=1, max_spare_cursors=32,
             max_spare_iters=32):
         envpp = _ffi.new('MDB_env **')
 
@@ -622,10 +645,12 @@ class Environment(object):
             flags |= MDB_NOSYNC
         if map_async:
             flags |= MDB_MAPASYNC
+        if not readahead:
+            flags |= MDB_NORDAHEAD
         if writemap:
             flags |= MDB_WRITEMAP
 
-        if isinstance(path, type(u'')):
+        if isinstance(path, UnicodeType):
             path = path.encode(sys.getfilesystemencoding())
 
         rc = mdb_env_open(self._env, path, flags, mode)
@@ -786,7 +811,7 @@ class Environment(object):
             rc = mdb_reader_list(self._env, _msg_func, _ffi.NULL)
             if rc:
                 raise _error("mdb_reader_list", rc)
-            return ''.join(_callbacks.msg_func)
+            return UnicodeType().join(_callbacks.msg_func)
         finally:
             del _callbacks.msg_func
 
@@ -821,7 +846,7 @@ class Environment(object):
 
             `name`:
                 Database name. If ``None``, indicates the main database should
-                be returned, otherwise indicates a sub-database should be
+                be returned, otherwise indicates a named database should be
                 created inside the main database. In other words, **a key
                 representing the database will be visible in the main database,
                 and the database name cannot conflict with any existing key**
@@ -849,6 +874,9 @@ class Environment(object):
                 If ``True``, create the database if it doesn't exist, otherwise
                 raise an exception.
         """
+        if isinstance(name, UnicodeType):
+            name = name.encode(sys.getfilesystemencoding())
+
         ref = self._dbs.get(name)
         if ref:
             db = ref()
@@ -867,78 +895,6 @@ class Environment(object):
         """Shortcut for :py:class:`lmdb.Transaction`"""
         return Transaction(self, db, parent, write, buffers)
 
-    def get(self, key, default=None, db=None):
-        """Use a temporary read transaction to invoke
-        :py:meth:`Transaction.get`."""
-        with Transaction(self) as txn:
-            return txn.get(key, default, db)
-
-    def gets(self, keys, db=None):
-        """Use a temporary read transaction to invoke
-        :py:meth:`Transaction.get` for each key in `keys`. The returned value
-        is a dict containing one element for each key that existed."""
-        dct = {}
-        with Transaction(self) as txn:
-            for key in keys:
-                value = txn.get(key, None, db)
-                if value is not None:
-                    dct[key] = value
-        return dct
-
-    def put(self, key, value, dupdata=False, overwrite=True, append=False,
-            db=None):
-        """Use a temporary write transaction to invoke
-        :py:meth:`Transaction.put`."""
-        with Transaction(self, write=True) as txn:
-            txn.put(key, value, dupdata, overwrite, append, db)
-
-    def puts(self, items, dupdata=False, overwrite=True, append=False,
-             db=None):
-        """Use a temporary write transaction to invoke
-        :py:meth:`Transaction.put` as `put(x, y)` for each `(x, y)`
-        in `items`. Items must be a dict, or an iterable producing 2-tuples.
-        This function requires 2-tuples, no other sequence type is accepted.
-
-        Returns a list of :py:meth:`Transaction.put` return values.
-
-            .. code-block:: python
-
-                a_existed, b_existed = env.puts(overwrite=False, items={
-                    'a': '1',
-                    'b': '2'
-                })
-
-                if a_existed:
-                    print 'Did not overwrite a, it already existed.'
-                if b_existed:
-                    print 'Did not overwrite b, it already existed.'
-        """
-        if type(items) is dict:
-            items = items.iteritems()
-        with Transaction(self, write=True) as txn:
-            return [txn.put(key, value, dupdata, overwrite, append, db)
-                    for key, value in items]
-
-    def delete(self, key, value='', db=None):
-        """Use a temporary write transaction to invoke
-        :py:meth:`Transaction.delete`."""
-        with Transaction(self, write=True) as txn:
-            return txn.delete(key, value, db)
-
-    def deletes(self, keys, db=None):
-        """Use a temporary write transaction to invoke
-        :py:meth:`Transaction.delete` for each key in `keys`. Returns a list of
-        :py:meth:`Transaction.delete` return values."""
-        with Transaction(self, write=True) as txn:
-            return [txn.delete(key, '', db) for key in keys]
-
-    def cursor(self, buffers=False, db=None):
-        """Use a temporary read transaction to return a :py:class:`Cursor`. The
-        transaction will remain alive for as long as the cursor exists.
-        """
-        txn = Transaction(self, db, None, False, buffers)
-        return Cursor(db or self._db, txn)
-
 
 class _Database(object):
     """Internal database handle."""
@@ -995,9 +951,9 @@ class Transaction(object):
             Environment the transaction should be on.
 
         `db`:
-            Default sub-database to operate on. If unspecified, defaults to the
-            environment's main database. Can be overridden on a per-call basis
-            below.
+            Default named database to operate on. If unspecified, defaults to
+            the environment's main database. Can be overridden on a per-call
+            basis below.
 
         `parent`:
             ``None``, or a parent transaction (see lmdb.h).
@@ -1044,6 +1000,9 @@ class Transaction(object):
         if rc:
             raise _error("mdb_txn_begin", rc)
         self._txn = txnpp[0]
+        # Number of mutations occurred since start of transaction. Required to
+        # know when Cursor key/value must be refreshed.
+        self._mutations = 0
 
     def _invalidate(self):
         self.abort()
@@ -1075,8 +1034,8 @@ class Transaction(object):
         return self.env._convert_stat(st)
 
     def drop(self, db, delete=True):
-        """Delete all keys in a sub-database, and optionally delete the
-        sub-database itself. Deleting the sub-database causes it to become
+        """Delete all keys in a named database and optionally delete the named
+        database itself. Deleting the named database causes it to become
         unavailable, and invalidates existing cursors.
 
         Equivalent to `mdb_drop()
@@ -1084,6 +1043,7 @@ class Transaction(object):
         """
         _kill_dependents(db)
         rc = mdb_drop(self._txn, db._dbi, delete)
+        self._mutations += 1
         if rc:
             raise _error("mdb_drop", rc)
 
@@ -1155,6 +1115,10 @@ class Transaction(object):
                 If ``True``, append the pair to the end of the database without
                 comparing its order first. Appending a key that is not greater
                 than the highest existing key will cause corruption.
+
+            `db`:
+                Named database to operate on. If unspecified, defaults to the
+                database given to the :py:class:`Transaction` constructor.
         """
         flags = 0
         if not dupdata:
@@ -1166,13 +1130,32 @@ class Transaction(object):
 
         rc = pymdb_put(self._txn, (db or self._db)._dbi,
                        key, len(key), value, len(value), flags)
+        self._mutations += 1
         if rc:
             if rc == MDB_KEYEXIST:
                 return False
             raise _error("mdb_put", rc)
         return True
 
-    def delete(self, key, value='', db=None):
+    def replace(self, key, value, db=None):
+        """Use a temporary cursor to invoke :py:meth:`Cursor.replace`.
+
+            `db`:
+                Named database to operate on. If unspecified, defaults to the
+                database given to the :py:class:`Transaction` constructor.
+        """
+        return Cursor(db or self._db, self).replace(key, value)
+
+    def pop(self, key, db=None):
+        """Use a temporary cursor to invoke :py:meth:`Cursor.pop`.
+
+            `db`:
+                Named database to operate on. If unspecified, defaults to the
+                database given to the :py:class:`Transaction` constructor.
+        """
+        return Cursor(db or self._db, self).pop(key)
+
+    def delete(self, key, value=b'', db=None):
         """Delete a key from the database.
 
         Equivalent to `mdb_del()
@@ -1190,6 +1173,7 @@ class Transaction(object):
         """
         rc = pymdb_del(self._txn, (db or self._db)._dbi,
                        key, len(key), value, len(value))
+        self._mutations += 1
         if rc:
             if rc == MDB_NOTFOUND:
                 return False
@@ -1231,12 +1215,6 @@ class Cursor(object):
     meaning strange behavior results when multiple iterators exist on the same
     cursor.
 
-        ::
-
-            >>> with env.begin() as txn:
-            ...     for i, (key, value) in enumerate(txn.cursor().iterprev()):
-            ...         print '%dth last item is (%r, %r)' % (1 + i, key, value)
-
     Both :py:meth:`iternext` and :py:meth:`iterprev` accept `keys` and `values`
     arguments. If both are ``True``, then the value of :py:meth:`item` is
     yielded on each iteration. If only `keys` is ``True``, :py:meth:`key` is
@@ -1287,6 +1265,9 @@ class Cursor(object):
         if rc:
             raise _error("mdb_cursor_open", rc)
         self._cur = curpp[0]
+        # If Transaction.mutations!=last_mutation, must MDB_GET_CURRENT to
+        # refresh `key' and `val'.
+        self._last_mutation = txn._mutations
 
     def _invalidate(self):
         if self._cur:
@@ -1303,14 +1284,23 @@ class Cursor(object):
 
     def key(self):
         """Return the current key."""
+        # Must refresh `key` and `val` following mutation.
+        if self._last_mutation != self.txn._mutations:
+            self._cursor_get(MDB_GET_CURRENT)
         return self._to_py(self._key)
 
     def value(self):
         """Return the current value."""
+        # Must refresh `key` and `val` following mutation.
+        if self._last_mutation != self.txn._mutations:
+            self._cursor_get(MDB_GET_CURRENT)
         return self._to_py(self._val)
 
     def item(self):
         """Return the current `(key, value)` pair."""
+        # Must refresh `key` and `val` following mutation.
+        if self._last_mutation != self.txn._mutations:
+            self._cursor_get(MDB_GET_CURRENT)
         return self._to_py(self._key), self._to_py(self._val)
 
     def _iter(self, op, keys, values):
@@ -1358,6 +1348,12 @@ class Cursor(object):
 
         If the cursor was not yet positioned, it is moved to the last record in
         the database, otherwise iteration proceeds from the current position.
+
+        ::
+
+            >>> with env.begin() as txn:
+            ...     for i, (key, value) in enumerate(txn.cursor().iterprev()):
+            ...         print '%dth last item is (%r, %r)' % (1 + i, key, value)
         """
         if not self._valid:
             self.last()
@@ -1477,6 +1473,7 @@ class Cursor(object):
         v = self._valid
         if v:
             rc = mdb_cursor_del(self._cur, 0)
+            self.txn._mutations += 1
             if rc:
                 raise _error("mdb_cursor_del", rc)
             self._cursor_get(MDB_GET_CURRENT)
@@ -1532,6 +1529,7 @@ class Cursor(object):
             flags |= MDB_APPEND
 
         rc = pymdb_cursor_put(self._cur, key, len(key), val, len(val), flags)
+        self.txn._mutations += 1
         if rc:
             if rc == MDB_KEYEXIST:
                 return False
@@ -1539,6 +1537,56 @@ class Cursor(object):
         self._cursor_get(MDB_GET_CURRENT)
         return True
 
+    def replace(self, key, val):
+        """Store a record, returning its previous value if one existed. Returns
+        ``None`` if no previous value existed.
+
+        This uses the best available mechanism to minimize the cost of a
+        `set-and-return-previous` operation.
+
+            `key`:
+                String key to store.
+
+            `value`:
+                String value to store.
+        """
+        flags = MDB_NOOVERWRITE
+        keylen = len(key)
+        rc = pymdb_cursor_put(self._cur, key, keylen, val, len(val), flags)
+        self.txn._mutations += 1
+        if not rc:
+            return
+        if rc != MDB_KEYEXIST:
+            raise _error("mdb_cursor_put", rc)
+
+        self._cursor_get(MDB_GET_CURRENT)
+        old = _mvstr(self._val)
+        rc = pymdb_cursor_put(self._cur, key, keylen, val, len(val), 0)
+        self.txn._mutations += 1
+        if rc:
+            raise _error("mdb_cursor_put", rc)
+        self._cursor_get(MDB_GET_CURRENT)
+        return old
+
+    def pop(self, key):
+        """Fetch a record's value then delete it. Returns ``None`` if no
+        previous value existed.
+
+        This uses the best available mechanism to minimize the cost of a
+        `delete-and-return-previous` operation.
+
+            `key`:
+                String key to delete.
+        """
+        if self._cursor_get_key(MDB_SET_KEY, key):
+            old = _mvstr(self._val)
+            rc = mdb_cursor_del(self._cur, 0)
+            self.txn._mutations += 1
+            if rc:
+                raise _error("mdb_cursor_del", rc)
+            self._cursor_get(MDB_GET_CURRENT)
+            return old
+
     def _iter_from(self, k, reverse):
         """Helper for centidb. Please do not rely on this interface, it may be
         removed in future.
